# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for converters."""

import datetime
import pyglove as pg
from vizier import pyvizier as vz
from vizier._src.pyglove import constants
from vizier._src.pyglove import converters
from absl.testing import absltest
from absl.testing import parameterized


class VizierCreatedSearchSpaceTest(parameterized.TestCase):
  """Tests for search space created from vz.SearchSpace."""

  def _search_space(self) -> vz.SearchSpace:
    ss = vz.SearchSpace()
    root = ss.root
    root.add_float_param('double', -1., 1.)
    root.add_int_param('int', 2, 4)
    root.add_discrete_param('discrete_int', (-2, 4, 7, 8))
    root.add_discrete_param('discrete_double', (-2.1, 4.1, 7.1, 8.1))
    root.add_categorical_param('categorical', ('a', 'b', 'c'))
    return ss

  def test_dna_spec(self):
    vc = converters.VizierConverter.from_problem(
        vz.ProblemStatement(self._search_space()))
    self.assertLen(vc.dna_spec, 5)
    self.assertNotEmpty(
        vc.problem.metadata.ns(constants.METADATA_NAMESPACE)[
            constants.STUDY_METADATA_KEY_DNA_SPEC])
    self.assertEqual(
        vc.problem.metadata.ns(constants.METADATA_NAMESPACE)[
            constants.STUDY_METADATA_KEY_USES_EXTERNAL_DNA_SPEC
        ],
        'false',
    )

  def test_dna_to_trial(self):
    vc = converters.VizierConverter.from_problem(
        vz.ProblemStatement(self._search_space()))
    self.assertDictEqual(
        vc.to_trial(vc.dna_spec.first_dna(),
                    fallback='raise_error').parameters.as_dict(), {
                        'double': -1,
                        'int': 2,
                        'discrete_int': -2,
                        'discrete_double': -2.1,
                        'categorical': 'a'
                    })

  def test_vizier_trial_to_tuner_trial(self):
    vc = converters.VizierConverter.from_problem(
        vz.ProblemStatement(self._search_space())
    )
    trial = vz.Trial()
    trial.id = 1
    trial.description = 'A test trial'
    trial.parameters['double'] = -0.5
    trial.parameters['int'] = 3
    trial.parameters['discrete_int'] = 7
    trial.parameters['discrete_double'] = 4.1
    trial.parameters['categorical'] = 'a'
    self.assertEqual(
        vc.to_tuner_trial(trial),
        pg.tuning.Trial(
            id=1,
            description=trial.description,
            dna=pg.DNA([-0.5, 1, 2, 1, 0]),
            related_links={},
            measurements=[],
            final_measurement=None,
            status='PENDING',
            created_time=int(  # pylint: disable=g-long-ternary
                trial.creation_time.replace(
                    tzinfo=datetime.timezone.utc
                ).timestamp()
            )
            if trial.creation_time
            else None,
            completed_time=None,
        ),
    )

  @parameterized.parameters((7, None), (7.0, 'pyglove'))
  def test_trial_to_dna(self, discrete_int_value, metadata_ns):
    vc = converters.VizierConverter.from_problem(
        vz.ProblemStatement(search_space=self._search_space()))
    trial = vz.Trial()
    trial.parameters['double'] = -.5
    trial.parameters['int'] = 3
    trial.parameters['discrete_int'] = discrete_int_value
    trial.parameters['discrete_double'] = 4.1
    trial.parameters['categorical'] = 'a'

    metadata = trial.metadata.ns(metadata_ns) if metadata_ns else trial.metadata
    metadata[constants.TRIAL_METADATA_KEY_DNA_METADATA] = '{"log_prob": 1.0}'
    dna = vc.to_dna(trial)
    self.assertEqual(dna, pg.DNA([-0.5, 1, 2, 1, 0]))
    self.assertEqual(dna.metadata, dict(log_prob=1.0))


class PyGloveCreatedSearchSpaceTest(parameterized.TestCase):
  """Tests for PyGlove created search space."""

  def _dna_spec(self) -> pg.DNASpec:

    class CustomTypeDecision(pg.hyper.CustomHyper):

      def custom_decode(self, dna):
        return dna.value

    search_space = pg.Dict(
        a=CustomTypeDecision(),
        b=pg.manyof(2, [0, 0, 2, 2]),
        x=pg.oneof([3, 2]),
        y=pg.floatv(0.1, 2.0),
        z=pg.oneof(['foo', 'bar']),
    )
    return pg.dna_spec(search_space)

  def test_search_space(self):
    vc = converters.VizierConverter.from_dna_spec(self._dna_spec())
    # Custom-type decision point `a` is not part of the search space.
    # But `b` has two decision points.
    self.assertLen(vc.search_space.parameters, 5)
    self.assertNotEmpty(
        vc.problem.metadata.ns(constants.METADATA_NAMESPACE)[
            constants.STUDY_METADATA_KEY_DNA_SPEC])
    self.assertEqual(
        vc.problem.metadata.ns(constants.METADATA_NAMESPACE)[
            constants.STUDY_METADATA_KEY_USES_EXTERNAL_DNA_SPEC
        ],
        'true',
    )

  def test_dna_to_trial(self):
    dna_spec = self._dna_spec()
    vc = converters.VizierConverter.from_dna_spec(dna_spec)
    dna = pg.DNA(['abc', [1, 2], 0, 0.5, 1], spec=dna_spec)
    trial = vc.to_trial(dna, fallback='raise_error')
    self.assertLen(trial.parameters, 5)
    self.assertEqual(trial.parameters['b[0]'], vz.ParameterValue('1/4 (0)'))
    self.assertEqual(trial.parameters['b[1]'], vz.ParameterValue('2/4 (2)'))
    self.assertEqual(trial.parameters['x'], vz.ParameterValue(3))
    self.assertEqual(trial.parameters['y'], vz.ParameterValue(0.5))
    self.assertEqual(trial.parameters['z'], vz.ParameterValue('1/2 (\'bar\')'))
    self.assertEqual(
        pg.from_json_str(
            trial.metadata.ns(constants.METADATA_NAMESPACE)
            [constants.TRIAL_METADATA_KEY_CUSTOM_TYPE_DECISIONS]), {'a': 'abc'})

  def test_trial_to_dna(self):
    vc = converters.VizierConverter.from_dna_spec(self._dna_spec())
    trial = vz.Trial()
    trial.parameters['x'] = 2.0
    trial.parameters['b[0]'] = '1/4 (0)'
    trial.parameters['b[1]'] = '2/4 (2)'
    trial.parameters['y'] = 0.5
    trial.parameters['z'] = '1/2 (\'bar\')'
    trial.metadata.ns(constants.METADATA_NAMESPACE)[
        constants.TRIAL_METADATA_KEY_CUSTOM_TYPE_DECISIONS] = pg.to_json_str(
            {'a': 'abc'})
    self.assertEqual(vc.to_dna(trial), pg.DNA(['abc', [1, 2], 1, 0.5, 1]))

  def test_dna_trial_conversion(self):
    dna_spec_str = (
        '/Td6WFoAAATm1rRGAgAhARYAAAB0L+Wj4DRRAd1dAD2Ih+dECUOcj7RXvfBPYS0OyzAc2w'
        '+PsBIJ1+EKYkDTqQ48R4Lre1bjAsSR5xLCA3KFq2tR8zHqH2SfrVoBUCrdm1gXQLM7PnGC'
        'z52uZFScSqVSvmJfgDEysdKpt//4+0S1c3hId0yS85IVlpXp68vfx+7pUOGCrns7aqQMnQ'
        'lU6+71UcejsPe7GyuCOnl1hRUMWBJ4TlvES+DZG5v9bDPB84v3vtEMuj3e6qGl67y1Un6F'
        '/P2g9qXeBt7DO2zcGdFjdJbkhdQBAFM/+5YsTXP7N5hy6Ih0FYlLUh0SoALlEwGfqyzR1D'
        'QSyZWoCYilaf67UqdgusN6+RSJ8CXOgNPcMZEUwCd1gvzqOKlqTKu4tELJWh/vhO4pWGmQ'
        'T/DC6xlS9/6s0o1nzYO5hOQwOXmZ7G+VTuxi7k267ALMIsn197fMQVnSjNCVCCr1vkheyV'
        'p9wLygytvhWS0lXtlzNGFQMB3z/62OC6A3dz4hM52CACxaHyjS3WtYOGNUDHoMGew0wiSj'
        '/OIAlJ7aQXaz5kXtH8OsfqXTOWecW0DnhtWtKg8nolR0rqU7y6uoPWX+7HQYnzdZDOeoSK'
        'ZtDCKMRUam52iceU9o+1o4SB6XtWsO3JkOQ+5ZTwZ6zgqQAAAAAAB4E+94/gOmcgAB+QPS'
        'aAAAVg3xfLHEZ/sCAAAAAARZWg=='
    )
    dna_spec = converters.restore_dna_spec(dna_spec_str)
    converter = converters.VizierConverter.from_dna_spec(dna_spec)
    dna = pg.DNA(
        [
            2,
            1,
            [3, 4],
            1,
            0,
            [4, 5],
            1,
            [1, 2],
            0,
            [2, 3],
            0,
            [0, 4],
            0,
            [1, 9],
            0,
            [2, 5],
            [4, 1, 3, 2, 0],
        ],
        spec=dna_spec,
    )
    trial = converter.to_trial(dna, fallback='raise_error')
    self.assertTrue(pg.eq(converter.to_dna(trial), dna))


class GetPyGloveMetadataTest(absltest.TestCase):

  def test_metadata_from_pyglove_namespace(self):
    """Tests getting metadata from PyGlove namespace."""
    t = vz.Trial()
    t.metadata.ns(constants.METADATA_NAMESPACE)['foo'] = '1.0'
    self.assertEqual(converters.get_pyglove_metadata(t), {'foo': 1.0})

  def test_metadata_from_empty_namespace(self):
    """Tests getting metadata from empty namespace."""
    t = vz.Trial()
    t.metadata['DNA_METADATA'] = '{"log_prob": 1.0}'
    t.metadata['should_exclude'] = 'abc'
    self.assertEqual(
        converters.get_pyglove_metadata(t), {'DNA_METADATA': {'log_prob': 1.0}}
    )


class MakeParameterConfigTest(absltest.TestCase):
  """Tests for make parameter configs from DNASpec."""

  def testScale(self):
    """Test scale type conversion."""
    self.assertEqual(converters.get_scale_type(None), vz.ScaleType.LINEAR)
    self.assertEqual(converters.get_scale_type('linear'), vz.ScaleType.LINEAR)
    self.assertEqual(converters.get_scale_type('log'), vz.ScaleType.LOG)
    self.assertEqual(
        converters.get_scale_type('rlog'), vz.ScaleType.REVERSE_LOG)
    with self.assertRaisesRegex(
        ValueError, 'Unsupported scale type'):
      converters.get_scale_type('unknown')

  def testHyperValueAsRoot(self):
    """Test using hyper value as search space root."""
    search_space = pg.oneof([1, 2])
    actual = converters.VizierConverter.from_dna_spec(
        pg.dna_spec(search_space)).search_space
    expected = vz.SearchSpace()
    expected.root.add_discrete_param('$', [1, 2])
    self.assertEqual(actual, expected)

  def testFlatSearchSpace(self):
    """Test flat search spaces."""

    class CustomDecisionPoint(pg.geno.CustomDecisionPoint):

      def custom_decode(self, dna):
        return dna.value

    search_space = pg.Dict(
        # custom decision point 'a' will not be inserted as a part of
        # parameter config.
        a=CustomDecisionPoint(),
        b=pg.oneof([2, 1]),
        x=pg.oneof([2, 2, 1]),
        y=[pg.floatv(1e-6, 1.0, scale='log')],
        z=pg.Dict(p=pg.manyof(2, ['foo', 'bar'])),
    )
    actual = converters.VizierConverter.from_dna_spec(
        pg.dna_spec(search_space)).search_space
    expected = vz.SearchSpace()
    root = expected.root
    # Feasible points of discrete params are sorted.
    root.add_discrete_param('b', [1, 2])
    root.add_categorical_param('x', ['0/3 (2)', '1/3 (2)', '2/3 (1)'])
    root.add_float_param('y[0]', 1e-6, 1.0, scale_type=vz.ScaleType.LOG)
    root.add_categorical_param('z.p[0]', ['0/2 (\'foo\')', '1/2 (\'bar\')'])
    root.add_categorical_param('z.p[1]', ['0/2 (\'foo\')', '1/2 (\'bar\')'])

    self.assertEqual(actual, expected)

  def testHierarchicalSearchSpace(self):
    """Test hierarchical search space."""
    search_space = pg.oneof([[pg.oneof([1, 2])],
                             pg.Dict(x=pg.floatv(0.1, 0.5)),
                             pg.manyof(2,
                                       [pg.oneof(['foo', 'bar']), True, False])
                            ])
    actual = converters.VizierConverter.from_dna_spec(
        pg.dna_spec(search_space)).search_space

    expected = vz.SearchSpace()
    root = expected.root
    categories = [
        '0/3 ([0: OneOf(candidates=[0: 1, 1: 2])])',
        '1/3 ({x=Float(min_value=0.1, max_value=0.5)})',
        ('2/3 (ManyOf(num_choices=2, candidates=[0: OneOf('
         'candidates=[0: \'foo\', 1: \'bar\']), 1: True, 2: False]))')
    ]
    dollar = root.add_categorical_param('$', categories)
    dollar.select_values([categories[0]]).add_discrete_param(
        '[=0/3][0]',
        [
            1,
            2,
        ],
    )
    dollar.select_values([categories[1]]).add_float_param(
        name='[=1/3].x',
        min_value=0.1,
        max_value=0.5,
        scale_type=vz.ScaleType.LINEAR,
    )

    branch2 = dollar.select_values([categories[2]])

    def _add_children(name):
      branch2.add_categorical_param(
          name,
          [
              '0/3 (OneOf(candidates=[0: \'foo\', 1: \'bar\']))', '1/3 (True)',
              '2/3 (False)'
          ],
      ).select_values(['0/3 (OneOf(candidates=[0: \'foo\', 1: \'bar\']))'
                      ]).add_categorical_param(
                          f'{name}[=0/3]',
                          [
                              '0/2 (\'foo\')',
                              '1/2 (\'bar\')',
                          ],
                      )

    _add_children('[=2/3][0]')
    _add_children('[=2/3][1]')
    self.assertEqual(
        expected, actual, msg=f'expected={expected}, actual={actual}')


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Wrapper around optproblems PyPI library for multiobjective problems.

See documentation:
https://ls11-www.cs.tu-dortmund.de/people/swessing/optproblems/doc/index.html#

WFG benchmark was used in MultiObjective TPE:
https://www.jair.org/index.php/jair/article/view/13188
"""

import json
import attrs
from optproblems import dtlz
from optproblems import wfg
from optproblems import zdt
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter_factory
from vizier._src.benchmarks.experimenters import numpy_experimenter


def _DefaultOptProblemStatement(
    dimension: int,
    num_objectives: int,
    *,
    min_value: float = 0.0,
    max_value: float = 1.0,
):
  """Returns default optproblems ProblemStatement."""
  problem = vz.ProblemStatement()
  for n in range(num_objectives):
    metric = vz.MetricInformation(
        name=f"f{n}", goal=vz.ObjectiveMetricGoal.MINIMIZE
    )
    problem.metric_information.append(metric)
  for d in range(dimension):
    problem.search_space.root.add_float_param(f"x{d}", min_value, max_value)
  return problem


EXPERIMENTER_FACTORY_KEY = "experimenter_factory"


@attrs.define
class WFGExperimenterFactory(
    experimenter_factory.SerializableExperimenterFactory
):
  """Multiobjective problem with variable number of objectives and dimensions.

  Reference:
    Huband, S.; Hingston, P.; Barone, L.; While, L. (2006). A review of
    multiobjective test problems and a scalable test problem toolkit. IEEE
    Transactions on Evolutionary Computation, vol.10, no.5, pp. 477-506.
  """

  name: str = attrs.field(
      default="", validator=attrs.validators.instance_of(str)
  )
  dim: int = attrs.field(
      default=1,
      validator=[attrs.validators.instance_of(int), attrs.validators.gt(0)],
  )
  num_objectives: int = attrs.field(
      default=2,
      validator=[attrs.validators.instance_of(int), attrs.validators.ge(2)],
  )

  def __attrs_post_init__(self):
    # k = "Position-related parameters". Must be divisible by (num_obj-1).
    k = self.num_objectives - 1
    if (self.dim - k) % 2 != 0:
      raise ValueError(
          f"dimensions - k must be even, got {self.dim - k} for k={k}."
      )
    self.k = k

  def __call__(self) -> numpy_experimenter.MultiObjectiveNumpyExperimenter:
    optprob_factory = getattr(wfg, self.name, None)
    if optprob_factory is None:
      raise ValueError(f"{self.name} is not a valid WFG problem in wfg.py")
    optprob: wfg.WFGBaseProblem = optprob_factory(
        self.num_objectives, self.dim, self.k
    )
    impl = optprob.objective_function
    problem = _DefaultOptProblemStatement(self.dim, self.num_objectives)
    return numpy_experimenter.MultiObjectiveNumpyExperimenter(impl, problem)

  def dump(self) -> vz.Metadata:
    metadata = vz.Metadata()
    metadata_dict = {
        "name": self.name,
        "dim": self.dim,
        "num_objectives": self.num_objectives,
    }
    metadata[EXPERIMENTER_FACTORY_KEY] = json.dumps(metadata_dict)
    return metadata

  @classmethod
  def recover(cls, metadata: vz.Metadata) -> "WFGExperimenterFactory":
    metadata_dict = json.loads(metadata[EXPERIMENTER_FACTORY_KEY])
    return cls(**metadata_dict)


@attrs.define
class DTLZExperimenterFactory(
    experimenter_factory.SerializableExperimenterFactory
):
  """Multiobjective problem with variable number of objectives and dimensions.

  Reference:
    K. Deb, L. Thiele, M. Laumanns, E. Zitzler, A. Abraham, L. Jain, and
    R. Goldberg. Scalable test problems for evolutionary multi-objective
    optimization. Evolutionary Multiobjective Optimization, Springer-Verlag,
    pp. 105-145, 2005.
  """

  name: str = attrs.field(
      default="", validator=attrs.validators.instance_of(str)
  )
  dim: int = attrs.field(
      default=1,
      validator=[attrs.validators.instance_of(int), attrs.validators.gt(0)],
  )
  num_objectives: int = attrs.field(
      default=2,
      validator=[attrs.validators.instance_of(int), attrs.validators.ge(2)],
  )

  def __call__(self) -> numpy_experimenter.MultiObjectiveNumpyExperimenter:
    optprob_factory = getattr(dtlz, self.name, None)
    if optprob_factory is None:
      raise ValueError(f"{self.name} is not a valid DTLZ problem in dtlz.py")
    optprob: dtlz.DTLZBaseProblem = optprob_factory(
        self.num_objectives, self.dim
    )
    impl = optprob.objective_function
    problem = _DefaultOptProblemStatement(self.dim, self.num_objectives)
    return numpy_experimenter.MultiObjectiveNumpyExperimenter(impl, problem)

  def dump(self) -> vz.Metadata:
    metadata = vz.Metadata()
    metadata_dict = {
        "name": self.name,
        "dim": self.dim,
        "num_objectives": self.num_objectives,
    }
    metadata[EXPERIMENTER_FACTORY_KEY] = json.dumps(metadata_dict)
    return metadata

  @classmethod
  def recover(cls, metadata: vz.Metadata) -> "DTLZExperimenterFactory":
    metadata_dict = json.loads(metadata[EXPERIMENTER_FACTORY_KEY])
    return cls(**metadata_dict)


@attrs.define
class ZDTExperimenterFactory(
    experimenter_factory.SerializableExperimenterFactory
):
  """Multiobjective problem with two objectives and d-dimensions.

  Reference:
    E. Zitzler, K. Deb, and L. Thiele. Comparison of multiobjective
    evolutionary algorithms: Empirical results. Evolutionary Computation, vol.
    8, no. 2,pp. 173-195, 2000.
  """

  name: str = attrs.field(
      default="", validator=attrs.validators.instance_of(str)
  )
  dim: int = attrs.field(
      default=1,
      validator=[attrs.validators.instance_of(int), attrs.validators.gt(0)],
  )

  def __call__(self) -> numpy_experimenter.MultiObjectiveNumpyExperimenter:
    optprob_factory = getattr(zdt, self.name, None)
    if optprob_factory is None:
      raise ValueError(f"{self.name} is not a valid ZDT problem in zdt.py")
    if optprob_factory == zdt.ZDT5:
      raise ValueError("ZDT5 does not allow variable dimensions.")
    optprob: zdt.ZDTBaseProblem = optprob_factory(self.dim)
    impl = optprob.objective_function
    problem = _DefaultOptProblemStatement(self.dim, 2)
    return numpy_experimenter.MultiObjectiveNumpyExperimenter(impl, problem)

  def dump(self) -> vz.Metadata:
    metadata = vz.Metadata()
    metadata_dict = {
        "name": self.name,
        "dim": self.dim,
    }
    metadata[EXPERIMENTER_FACTORY_KEY] = json.dumps(metadata_dict)
    return metadata

  @classmethod
  def recover(cls, metadata: vz.Metadata) -> "ZDTExperimenterFactory":
    metadata_dict = json.loads(metadata[EXPERIMENTER_FACTORY_KEY])
    return cls(**metadata_dict)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Init file."""
import os
import sys

THIS_DIR = os.path.dirname(os.path.realpath(__file__))
PROTO_ROOT = os.path.realpath(os.path.join(THIS_DIR, "_src", "service"))

sys.path.append(PROTO_ROOT)

__version__ = "0.1.16"



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter that permutes the parameters before evaluation."""

import copy
import logging
from typing import Optional, Sequence

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter


class PermutingExperimenter(experimenter.Experimenter):
  """PermutingExperimenter permutes discrete/categorical parameters."""

  def __init__(
      self,
      exptr: experimenter.Experimenter,
      parameters_to_permute: Sequence[str],
      seed: Optional[int] = None,
  ):
    """PermutingExperiment permutes discrete parameter values before passing to exptr.

    Args:
      exptr: Underlying experimenter to be wrapped.
      parameters_to_permute: Parameter names that are to be permuted.
      seed:

    Raises:
      ValueError: Condtional search space.
    """
    self._exptr = exptr
    exptr_problem_statement = exptr.problem_statement()
    self._problem_statement = exptr_problem_statement

    self._rng = np.random.default_rng(seed)
    if exptr_problem_statement.search_space.is_conditional:
      raise ValueError(
          'Search space should not have conditional'
          f' parameters  {exptr_problem_statement}'
      )

    self._parameter_permutation_dict = {}
    for parameter_name in parameters_to_permute:
      parameter = exptr_problem_statement.search_space.get(parameter_name)
      if not np.isfinite(parameter.num_feasible_values):
        raise ValueError(
            f'Parameter to permute {parameter} is continuous.'
            ' Permuting continuous parameters is not supported.'
        )

      permutation_list = self._rng.permuted(parameter.feasible_values)
      permutation_dict = {
          a: b for a, b in zip(parameter.feasible_values, permutation_list)
      }
      self._parameter_permutation_dict[parameter.name] = permutation_dict

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]) -> None:
    """Evaluate the trials after permuting the parameters."""
    previous_parameters = [suggestion.parameters for suggestion in suggestions]
    self._permute(suggestions)
    self._exptr.evaluate(suggestions)
    # Replace stored previous parameters.
    for parameters, suggestion in zip(previous_parameters, suggestions):
      suggestion.parameters = parameters

  def _permute(self, suggestions: Sequence[pyvizier.Trial]) -> None:
    """Permutes parameter values in place."""
    for suggestion in suggestions:
      new_parameters = {}
      for name, parameter in suggestion.parameters.items():
        if name in self._parameter_permutation_dict:
          permutation_dict = self._parameter_permutation_dict[name]
          print('permutation dict', permutation_dict)
          logging.info('Permuting %s ', permutation_dict)
          new_parameters[name] = permutation_dict[parameter.value]
        else:
          new_parameters[name] = parameter
      suggestion.parameters = new_parameters

  def __repr__(self):
    return f'PermutingExperimenter on {str(self._exptr)}'



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Binary classifiers for Bayesian Optimization."""

from typing import Optional

import attr
import numpy as np
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.gaussian_process.kernels import ConstantKernel
from sklearn.gaussian_process.kernels import RBF

# TODO: Replace the sklearn GP classifier with TFP GP classifier
# once implemented.


@attr.define
class SklearnClassifier:
  """Class for Sklearn classifiers.

  Attributes:
    classifier: a sklearn classifier such as SVM and GaussianProcessClassifier.
    features: (n, d) shaped array of n samples in dimension d.
    labels: (n, 1) shaped array of binary labels in {0, 1}.
    features_test: (m, d) shaped array of m samples in dimension d.
    eval_metric: a string denoting the evaluation metric. Accepted options are
      `probability` which estimates the probability of belonging to each class
      and `decision` which estimates a non-probability based metric such as the
      margin from the classification boundary.
  """
  classifier: Optional[GaussianProcessClassifier] = attr.field(
      kw_only=True,
      default=GaussianProcessClassifier(
          kernel=ConstantKernel(1.) * RBF(length_scale=1.)))
  features: np.ndarray = attr.field(kw_only=True)
  labels: np.ndarray = attr.field(kw_only=True)
  features_test: np.ndarray = attr.field(kw_only=True)
  eval_metric: str = attr.field(kw_only=True, default='probability')

  def _check_features_and_labels_shapes(self) -> None:
    """Checks the compatibility between features and labels shapes."""
    if np.ndim(self.features) != 2:
      raise ValueError(f'{self} expects 2d features.')
    if self.labels.shape[0] != self.features.shape[0]:
      raise ValueError(f'There are `{self.features.shape[0]}` features and '
                       f'`{self.labels.shape[0]}` labels which is incompatible')
    if np.ndim(self.labels) != 1 and self.labels.shape[1] != 1:
      raise ValueError(
          f'{self} expects 1d labels or labels of shape (num_samples, 1), but'
          f'was given labels of shape `{self.labels.shape}` .')
    if self.features_test.shape[1] != self.features.shape[1]:
      raise ValueError(
          f'{self} features_test to have `{self.features.shape[1]}`,'
          f'but it has `{self.features_test.shape[1]}` features.')

  def _check_labels_values(self) -> None:
    if not set(self.labels).issubset({0, 1}):
      raise ValueError('Labels should be either zero or one.')
    if set(self.labels).issubset({0}) or set(self.labels).issubset({1}):
      raise ValueError(f'{self} expects at least one sample per class, but all'
                       'training labels contain the same class.')

  def _check_eval_metric(self) -> None:
    if self.eval_metric not in ['probability', 'decision']:
      raise ValueError(f'{self} expects the evaluation metric to be'
                       f'`probability` or `decision ` but `{self.eval_metric}`'
                       'was given.')

  # TODO: separate the training and evaluation for extra speed up.
  # Currently, the classifiers we use are reasonably fast.
  def __call__(self) -> np.ndarray:
    self._check_features_and_labels_shapes()
    self._check_labels_values()
    self._check_eval_metric()
    if self.classifier is None:
      raise RuntimeError('Classifier is None.')
    self.classifier.fit(np.asarray(self.features), np.asarray(self.labels))
    if self.eval_metric == 'probability':
      return self.classifier.predict_proba(np.asarray(self.features_test))[:, 1]
    else:
      return self.classifier.decision_function(np.asarray(self.features_test))  # pytype:disable=attribute-error



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""PyVizier module. See VizierConverter class."""

import base64
import datetime
import json
import lzma
import numbers
from typing import Any, Dict, List, Optional, Sequence, Literal

from absl import logging
import attr
import pyglove as pg
from vizier import pyvizier as vz
from vizier._src.pyglove import constants


def _to_json_str_compressed(value: Any) -> str:
  """Serialize (maybe) symbolic object to compressed JSON value."""
  return base64.b64encode(
      lzma.compress(json.dumps(
          pg.to_json(value)).encode('utf-8'))).decode('ascii')


def _parameter_with_external_type(
    val: vz.ParameterValueTypes,
    external_type: vz.ExternalType) -> vz.ParameterValueTypes:
  """Converts a parameter value to proper external type."""
  if external_type == vz.ExternalType.BOOLEAN:
    # We output strings 'True' or 'False', not booleans themselves.
    # because BOOLEAN is interally CATEGORICAL.
    return val
  elif external_type == vz.ExternalType.INTEGER:
    return int(val)
  elif external_type == vz.ExternalType.FLOAT:
    return float(val)
  else:
    return val


def _make_decision_point(
    parameter_config: vz.ParameterConfig) -> pg.geno.DecisionPoint:
  """Make a decision point (DNASpec) out from a parameter config."""

  # NOTE(daiyip): We set the name of each decision point instead of its
  # location with parameter name.
  #
  # Why? For conditional space, the ID of a decision point is a
  # path of locations from the root to the leaf node. For example, if there
  # are two parameters - a parent with location 'a' and a child with location
  # 'b', the ID for the child will be 'a.b'. However, for external (
  # non-PyGlove created) study, the parameter name for the child does not
  # follow this pattern. The solution is to use the ``name`` property of
  # `DNASpec`, which allows the user to access hierarchical decision
  # points by name, also DNA supports to_dict/from_dict based on the decision
  # point names instead of their IDs. Therefore, we can minimize the
  # difference between a PyGlove created study and an external study.

  name = parameter_config.name
  if parameter_config.type == vz.ParameterType.DOUBLE:
    # Create `pg.geno.Float` which does not have child spaces.
    min_value, max_value = parameter_config.bounds
    return pg.geno.Float(min_value, max_value, name=name)
  elif parameter_config.type in (vz.ParameterType.CATEGORICAL,
                                 vz.ParameterType.DISCRETE,
                                 vz.ParameterType.INTEGER):
    # Create `pg.geno.Choices` with possible child spaces.
    candidates = []
    literal_values = []
    for val in parameter_config.feasible_values:
      child_decision_points = []
      if parameter_config.child_parameter_configs:
        for child_pc in parameter_config.child_parameter_configs:
          if val in child_pc.matching_parent_values:
            child_decision_points.append(_make_decision_point(child_pc))
      candidates.append(pg.geno.Space(child_decision_points))
      literal_values.append(
          _parameter_with_external_type(val, parameter_config.external_type))
    return pg.geno.Choices(
        1, candidates, literal_values=literal_values, name=name)
  else:
    raise ValueError(
        f'Parameter Config Type {parameter_config.type!r} is not supported.')


def _to_dna_spec(search_space: vz.SearchSpace) -> pg.DNASpec:
  return pg.geno.Space(
      [_make_decision_point(pc) for pc in search_space.parameters])


def _to_search_space(dna_spec: pg.DNASpec) -> vz.SearchSpace:
  """Converts a DNASpec to Vizier search space.

  Args:
    dna_spec:

  Returns:
    Vizier search space.

  Raises:
    NotImplementedError: If no part of the spec can be converted to a Vizier
      parameter.
  """

  def _parameter_name(path: pg.KeyPath) -> str:
    # NOTE(daiyip): Vizier doesn't support empty name, thus we use a
    # special parameter name for the hyper value at root.
    return path.path if path else constants.PARAMETER_NAME_ROOT

  def _categories(spec: pg.geno.Choices) -> List[str]:
    return [spec.format_candidate(i) for i in range(len(spec.candidates))]

  def _category_value(spec: pg.geno.Choices, index: int) -> str:
    assert index < len(spec.candidates)
    return spec.format_candidate(index)

  def _add_dna_spec(root: vz.SearchSpaceSelector, path: pg.KeyPath,
                    spec: pg.DNASpec) -> None:
    """Convert a DNASpec node with parent choice to a list of parameters.

    Args:
      root: The DNA spec is added to this root.
      path: Root path of current DNA spec.
      spec: Current DNA spec.
    """
    if isinstance(spec, pg.geno.Space):
      for elem in spec.elements:
        _add_dna_spec(root, path + elem.location, elem)
    elif isinstance(spec, pg.geno.Choices):
      is_discrete = all(
          isinstance(v, numbers.Number) for v in spec.literal_values
      ) and len(set(spec.literal_values)) == len(spec.literal_values)

      for choice_idx in range(spec.num_choices):
        choice_path = path
        if spec.num_choices > 1:
          choice_path = choice_path + choice_idx

        if is_discrete:
          unique_feasible_points = sorted(set(spec.literal_values))
          root.add_discrete_param(
              name=_parameter_name(choice_path),
              # We sort the literal values since Vizier requires the feasible
              # points of a discrete parameter to be in increasing order.
              # The sorting has no impact to the trial parameter -> DNA
              # conversion since for numeric literal value, the conversion
              # is value based.
              feasible_values=unique_feasible_points,
          )
          if unique_feasible_points != spec.literal_values:
            logging.warning(
                'Candidates for parameter %r have been reordered/deduped from '
                '%s to %s to meet the sorted/distinct requirement for discrete '
                'parameter specifiction.',
                _parameter_name(choice_path),
                spec.literal_values,
                unique_feasible_points)
        else:
          new_parameter: vz.SearchSpaceSelector = root.add_categorical_param(
              name=_parameter_name(choice_path),
              feasible_values=_categories(spec))
          for candidate_idx, candidate in enumerate(spec.candidates):
            candidate_path = choice_path + pg.geno.ConditionalKey(
                candidate_idx, len(spec.candidates)
            )
            child: vz.SearchSpaceSelector = new_parameter.select_values(
                [_category_value(spec, candidate_idx)])
            _add_dna_spec(child, candidate_path, candidate)
    elif isinstance(spec, pg.geno.Float):
      root.add_float_param(
          name=_parameter_name(path),
          scale_type=get_scale_type(spec.scale),
          min_value=spec.min_value,
          max_value=spec.max_value)
    elif isinstance(spec, pg.geno.CustomDecisionPoint):
      # For CustomDecisionPoint, there is not a corresponding parameter type
      # in Vizier since its value is a variable string. In such case the
      # parameter value will be put into metadata.
      logging.info(
          'Encountered custom decision point %s, which will not be shown '
          'in Vizier dashboard.',
          _parameter_name(path),
      )
    else:
      raise NotImplementedError(
          f'Spec has unknown type. This Should never happen. Spec: {spec}')

  search_space = vz.SearchSpace()
  _add_dna_spec(search_space.root, pg.KeyPath(), dna_spec)

  if not search_space.parameters:
    raise NotImplementedError(
        'No part of the dna spec could be represented as a Vizier parameter.')
  return search_space


def get_scale_type(scale: Optional[str]) -> Optional[vz.ScaleType]:
  """Returns scale type based on scale string."""
  if scale in [None, 'linear']:
    return vz.ScaleType.LINEAR
  elif scale == 'log':
    return vz.ScaleType.LOG
  elif scale == 'rlog':
    return vz.ScaleType.REVERSE_LOG
  else:
    raise ValueError(f'Unsupported scale type: {scale!r}')


def get_pyglove_metadata(trial: vz.Trial) -> dict[str, Any]:
  """Extracts only the pyglove-related metadata into a simple dict."""
  metadata = dict()

  # NOTE(daiyip): This is to keep backward compatibility for Cloud NAS service,
  # which might loads trials from studies created in the old NAS pipeline for
  # transfer learning.
  for key, value in trial.metadata.items():
    if key in constants.TRIAL_METADATA_KEYS:
      metadata[key] = pg.from_json_str(value)

  for key, value in trial.metadata.ns(constants.METADATA_NAMESPACE).items():
    if key not in constants.TRIAL_METADATA_KEYS and value is not None:
      metadata[key] = pg.from_json_str(value)
  return metadata


def get_pyglove_study_metadata(problem: vz.ProblemStatement) -> pg.Dict:
  """Extracts only the pyglove-related metadata into a simple dict."""
  metadata = pg.Dict()
  pg_metadata = problem.metadata.ns(constants.METADATA_NAMESPACE)
  for key, value in pg_metadata.items():
    if key not in constants.STUDY_METADATA_KEYS and value is not None:
      metadata[key] = pg.from_json_str(value)
  return metadata


@attr.frozen
class VizierConverter:
  """Converts between PyGlove DNA and Vizier Trial.

  It can be initialized from a pg.DNASpec or vz.SearchSpace. It handles
  conversions between pg.DNA and vz.Trial.

  NOTE: Use a factory instead of __init__. There are two factories:
    VizierConverter.from_problem(...)
    VizierConverter.from_dna_spec(...)

  CAVEAT: The set of search spaces that can be described pg.DNASpec and
  vz.SearchSpace does not fully overlap. (e.g. DNASpec does not support
  discrete doubles and integers) As a result,
  `VizierConverter.from_dna_spec(VizierConverter.from_search_space(s).dna_spec)`
  is not the same as `VizierConverter.from_search_space(s)`.

  If VizierConverter is created from a DNA spec that can't be represented
  in Vizier search space, then `_problem` has a dummy search space and
  effectively has no purpose beyond saving dna_spec in metadata.
  In this case, `vizier_conversion_error` has a non-None value.
  """

  _dna_spec: pg.DNASpec = attr.field()
  _problem: vz.ProblemStatement = attr.field()
  _uses_external_dna_spec: bool = attr.field()

  vizier_conversion_error: Optional[Exception] = attr.field(
      default=None, kw_only=True)

  def __attrs_post_init__(self):
    # Store the dna spec in the metadata.
    self._problem.metadata.ns(constants.METADATA_NAMESPACE)[
        constants.STUDY_METADATA_KEY_DNA_SPEC] = _to_json_str_compressed(
            self._dna_spec)
    self._problem.metadata.ns(constants.METADATA_NAMESPACE)[
        constants.STUDY_METADATA_KEY_USES_EXTERNAL_DNA_SPEC
    ] = pg.to_json_str(self._uses_external_dna_spec)

  @property
  def metrics_to_optimize(self) -> Sequence[str]:
    metrics = []
    for m in self._problem.metric_information:
      if m.goal == vz.ObjectiveMetricGoal.MAXIMIZE:
        metrics.append(m.name)
      else:
        metrics.append(f'negative_{m.name}')
    return metrics

  @classmethod
  def from_problem(cls, problem: vz.ProblemStatement) -> 'VizierConverter':
    """Creates from vizier problem."""
    # TODO: Check this implementation
    json_str_compressed = problem.metadata.ns(constants.METADATA_NAMESPACE).get(
        constants.STUDY_METADATA_KEY_DNA_SPEC, None)

    if json_str_compressed is not None:
      dna_spec = restore_dna_spec(json_str_compressed)
    else:
      dna_spec = _to_dna_spec(problem.search_space)

    if bad_metrics := tuple(
        filter(lambda m: m.goal != vz.ObjectiveMetricGoal.MAXIMIZE,
               problem.metric_information)):
      raise ValueError(
          f'All goals must MAXIMIZE. Offending metrics: {bad_metrics}')
    if bad_metrics := tuple(
        filter(lambda m: m.type != vz.MetricType.OBJECTIVE,
               problem.metric_information)):
      logging.warning('All goals must be OBJECTIVE. Offending metrics: %s',
                      bad_metrics)

    uses_external_dna_spec = (
        json_str_compressed is not None
        and pg.from_json_str(
            problem.metadata.ns(constants.METADATA_NAMESPACE).get(
                constants.STUDY_METADATA_KEY_USES_EXTERNAL_DNA_SPEC, 'true'
            )
        )
    )
    return cls(dna_spec, problem, uses_external_dna_spec)

  @classmethod
  def from_dna_spec(
      cls,
      dna_spec: pg.DNASpec,
      metrics_to_maximize: Sequence[str] = (constants.REWARD_METRIC_NAME,)
  ) -> 'VizierConverter':
    """Create from dna spec."""
    problem = vz.ProblemStatement()
    for name in metrics_to_maximize:
      problem.metric_information.append(
          vz.MetricInformation(name, goal=vz.ObjectiveMetricGoal.MAXIMIZE))

    try:
      problem.search_space = _to_search_space(dna_spec)
      return cls(dna_spec, problem, True)
    except NotImplementedError as e:
      # Add a dummy parameter.
      problem.search_space.root.add_categorical_param(
          constants.DUMMY_PARAMETER_NAME,
          feasible_values=[constants.DUMMY_PARAMETER_VALUE])
      logging.info(
          'The provided DNA spec cannot be converted to a '
          'Vizier search space. Vizier algorithms cannot be used. '
          'Error was: %s', e)
      return cls(dna_spec, problem, True, vizier_conversion_error=e)

  @property
  def dna_spec(self) -> pg.DNASpec:
    return self._dna_spec

  @property
  def problem(self) -> vz.ProblemStatement:
    """Raises an error if the dna spec cannot be represented in Vizier."""
    if self.vizier_conversion_error:
      raise self.vizier_conversion_error
    return self._problem

  @property
  def uses_external_dna_spec(self) -> bool:
    """Returns True if the dna spec is provided from external."""
    return self._uses_external_dna_spec

  @property
  def problem_or_dummy(self) -> vz.ProblemStatement:
    """Returns dummy if the dna spec cannot be represented in Vizier."""
    return self._problem

  @property
  def search_space(self) -> vz.SearchSpace:
    if self.vizier_conversion_error:
      raise NotImplementedError(
          f'Vizier algorithms cannot work with the dna spec. '
          f'Error was: {self.vizier_conversion_error}')
    return self._problem.search_space

  def _process_key_value(self, key: str, value: vz.ParameterValueTypes):
    if key == constants.PARAMETER_NAME_ROOT:
      key = ''
    if self.dna_spec.hints == constants.FROM_VIZIER_STUDY_HINT:
      if not isinstance(value, str):
        value = float(value)  # Integers are always converted to doubles.
      if isinstance(self.dna_spec[key], pg.geno.Choices):
        value = repr(value)
    return key, value

  def _parameters_to_dict(self,
                          trial: vz.Trial) -> Dict[str, vz.ParameterValueTypes]:
    return dict({
        self._process_key_value(k, v)
        for k, v in trial.parameters.as_dict().items()
    })

  def to_dna(self, trial: vz.Trial) -> pg.DNA:
    """Extract DNA from vizier trial."""
    decision_dict = self._parameters_to_dict(trial)
    if len(
        decision_dict) == 1 and constants.DUMMY_PARAMETER_NAME in decision_dict:
      decision_dict = {}

    custom_decisions_str = trial.metadata.ns(constants.METADATA_NAMESPACE).get(
        constants.TRIAL_METADATA_KEY_CUSTOM_TYPE_DECISIONS, None)
    if custom_decisions_str is not None:
      custom_decisions = pg.from_json_str(custom_decisions_str)
      assert isinstance(custom_decisions, dict)
      decision_dict.update(custom_decisions)

    dna = pg.DNA.from_dict(
        decision_dict, self.dna_spec, use_ints_as_literals=True)

    # Restore DNA metadata if present
    dna_metadata = trial.metadata.ns(constants.METADATA_NAMESPACE).get(
        constants.TRIAL_METADATA_KEY_DNA_METADATA, None
    )

    if dna_metadata is None:
      # NOTE(daiyip): To be compatible with V1 pipeline for transfer learning,
      # we also try to read DNA_METADATA stored under the global (empty)
      # namespace.
      dna_metadata = trial.metadata.get(
          constants.TRIAL_METADATA_KEY_DNA_METADATA, None
      )

    if dna_metadata is not None:
      dna.rebind(
          metadata=pg.from_json_str(dna_metadata),
          skip_notification=True,
          raise_on_no_change=False,
      )
    else:
      logging.warn('DNA metadata is None for trial: %s', trial)
    return dna

  def to_trial(self, dna: pg.DNA, *,
               fallback: Literal['raise_error', 'return_dummy']) -> vz.Trial:
    """Converts DNA to vizier Trial.

    Args:
      dna:
      fallback: Decides the behavior when Vizier Search space is ill-formed from
        the DNA spec. 'raise_error': Raises an error. 'return_dummy': Returns a
        dummy trial.

    Returns:
      Vizier Trial.

    Raises:
      NotImplementedError: DNA has no valid representation as a Vizier trial.
    """
    trial = vz.Trial()
    trial.description = str(dna)
    trial.metadata.ns(constants.METADATA_NAMESPACE)[
        constants.TRIAL_METADATA_KEY_DNA_METADATA] = pg.to_json_str(
            dna.metadata)

    # Custom decision.
    def is_custom(x):
      return isinstance(x, pg.geno.CustomDecisionPoint)

    custom_decisions = dna.to_dict(filter_fn=is_custom)
    if custom_decisions:
      trial.metadata.ns(constants.METADATA_NAMESPACE)[
          constants.TRIAL_METADATA_KEY_CUSTOM_TYPE_DECISIONS] = pg.to_json_str(
              custom_decisions)

    if self.vizier_conversion_error:
      if fallback == 'return_dummy':
        trial.parameters[
            constants.DUMMY_PARAMETER_NAME] = constants.DUMMY_PARAMETER_VALUE
      else:
        raise self.vizier_conversion_error
    else:

      def is_discrete(x):
        return (
            isinstance(x, pg.geno.Choices)
            and all(isinstance(v, numbers.Number) for v in x.literal_values)
            and len(set(x.literal_values)) == len(x.literal_values)
        )

      if self._uses_external_dna_spec:
        key_type, value_type = 'id', 'choice_and_literal'
      else:
        key_type, value_type = 'name_or_id', 'literal'

      # Update (non-discrete) configured parameters from DNA.
      # We still write discrete parameter values in 'choice_and_literal' format
      # though their values will be overriden later, to preserve the order
      # of parameters.
      configured_parameters = dna.to_dict(
          key_type=key_type,
          value_type=value_type,
          filter_fn=lambda x: not is_custom(x),
      )

      # Update (discrete) configured parameters from DNA.
      # Force converting literal values to float to match DISCRETE parameter
      # expectation.
      discrete_parameters = {
          k: float(v)
          for k, v in dna.to_dict(
              key_type=key_type, value_type='literal', filter_fn=is_discrete
          ).items()
      }
      configured_parameters.update(discrete_parameters)

      if not configured_parameters:
        configured_parameters[constants.DUMMY_PARAMETER_NAME] = (
            constants.DUMMY_PARAMETER_VALUE
        )

      for name, value in configured_parameters.items():
        trial.parameters[name or constants.PARAMETER_NAME_ROOT] = (
            vz.ParameterValue(value)
        )
    return trial

  def to_tuner_measurement(
      self, vizier_measurement: Optional[vz.Measurement]
  ) -> Optional[pg.tuning.Measurement]:
    """Convert Vizier measurement to tuner Measurement."""
    if not vizier_measurement:
      return None
    tuner_measurement = pg.tuning.Measurement(
        step=int(vizier_measurement.steps),
        elapse_secs=vizier_measurement.elapsed_secs,
        reward=vizier_measurement.metrics.get_value(
            constants.REWARD_METRIC_NAME, 0.))
    tuner_measurement.metrics = {
        m: vizier_measurement.metrics.get_value(m, 0.)
        for m in vizier_measurement.metrics
    }
    return tuner_measurement

  def to_tuner_trial(self, vizier_trial: vz.Trial) -> pg.tuning.Trial:
    return pg.tuning.Trial(
        id=vizier_trial.id,
        description=vizier_trial.description,
        dna=self.to_dna(vizier_trial),
        metadata=dict(vizier_trial.metadata.ns(constants.METADATA_NAMESPACE)),
        related_links=dict(
            vizier_trial.metadata.ns(constants.METADATA_NAMESPACE).ns(
                constants.RELATED_LINKS_SUBNAMESPACE
            )
        ),
        measurements=vizier_trial.measurements,
        final_measurement=self.to_tuner_measurement(
            vizier_trial.final_measurement
        ),
        status=self._to_tuner_trial_status(vizier_trial.status),
        created_time=int(
            vizier_trial.creation_time.replace(
                tzinfo=datetime.timezone.utc
            ).timestamp()
        ),
        completed_time=int(  # pylint: disable=g-long-ternary
            vizier_trial.completion_time.replace(
                tzinfo=datetime.timezone.utc
            ).timestamp()
        )
        if vizier_trial.completion_time
        else None,
        infeasible=vizier_trial.infeasible,
    )

  def _to_tuner_trial_status(self, status: vz.TrialStatus) -> str:
    """Convert Vizier trial status to tuner trial status."""
    return 'PENDING' if status == vz.TrialStatus.ACTIVE else status.name


def restore_dna_spec(json_str_compressed: str) -> pg.DNASpec:
  """Restores DNASpec from compressed JSON str."""
  return pg.from_json(
      json.loads(lzma.decompress(base64.b64decode(json_str_compressed)))
  )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Harmonica algorithm for boolean search spaces from 'Hyperparameter Optimization: A Spectral Approach' (https://arxiv.org/abs/1706.00764).

This is a faithful re-implementation based off
https://github.com/callowbird/Harmonica.
"""
# pylint:disable=invalid-name
import abc
import itertools
from typing import Optional, Sequence, Set
import attrs
import numpy as np
from sklearn import linear_model
from sklearn import preprocessing

from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random


class _Surrogate(abc.ABC):
  """Utility class to organize code."""

  @abc.abstractmethod
  def reset(self) -> None:
    """Resets internal state."""

  @abc.abstractmethod
  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    """Performs regression and modifies internal state."""

  @abc.abstractmethod
  def predict(self, x: np.ndarray) -> float:
    """Performs prediction. Should NOT modify internal state."""


@attrs.define
class PolynomialSparseRecovery(_Surrogate):
  """Performs LASSO regression over low (d) degree polynomial coefficients."""

  # Maximum degree of monomials to use during polynomial regression.
  _d: int = attrs.field(
      default=3, validator=attrs.validators.ge(1), kw_only=True
  )
  # Number of top monomial coefficients to use.
  _num_top_monomials: int = attrs.field(
      default=5, validator=attrs.validators.ge(1), kw_only=True
  )

  # LASSO regularization coefficient.
  _alpha = attrs.field(
      default=3.0, validator=attrs.validators.ge(0.0), kw_only=True
  )

  # Internal state after performing regression.
  _poly_transformer: preprocessing.PolynomialFeatures = attrs.field(init=False)
  _top_poly_indices: np.ndarray = attrs.field(init=False)
  _top_poly_coefficients: np.ndarray = attrs.field(init=False)

  def __attrs_post_init__(self):
    self.reset()

  def reset(self) -> None:
    self._poly_transformer = preprocessing.PolynomialFeatures(
        self._d, interaction_only=True
    )
    self._top_poly_indices = np.empty(0)
    self._top_poly_coefficients = np.empty(0)

  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    """Performs LASSO regression to obtain top monomial coefficients."""

    # Computes monomial features for every vector in X.
    X_poly_features = self._poly_transformer.fit_transform(X)

    # Find optimial coefficients on monomials to the data.
    lasso_solver = linear_model.Lasso(fit_intercept=True, alpha=self._alpha)
    lasso_solver.fit(X_poly_features, Y)

    # Sort and obtain top coefficients.
    lasso_coefficients = lasso_solver.coef_
    poly_indices = np.argsort(-np.abs(lasso_coefficients))
    self._top_poly_indices = poly_indices[: self._num_top_monomials]
    self._top_poly_coefficients = lasso_coefficients[self._top_poly_indices]

  def predict(self, x: np.ndarray) -> float:
    """Predicts using polynomial features."""
    x_poly_features = self._poly_transformer.transform(x.reshape(1, -1))
    x_poly_features = x_poly_features[0]

    top_x_poly_features = np.take_along_axis(
        x_poly_features, self._top_poly_indices, axis=0
    )
    return np.dot(top_x_poly_features, self._top_poly_coefficients)

  def index_set(self) -> Set[int]:
    """Returns parameter index set J from top coefficients.

    For example, if our monomials corresponding to top cofficients were x0*x1,
    x3, x1*x2*x3, then the output would be Union({0, 1}, {3}, {1, 2, 3}).
    """
    indices = set()
    poly_feature_one_hots = self._poly_transformer.powers_
    for top_poly_index in self._top_poly_indices:
      active_indices = np.nonzero(poly_feature_one_hots[top_poly_index])
      active_indices = active_indices[0].tolist()
      indices = indices | set(active_indices)
    return indices


@attrs.define
class RestrictedSurrogate(_Surrogate):
  """New surrogate with input x's positions replaced from X_restrictor values."""

  X_restrictors: np.ndarray = attrs.field(init=True, kw_only=True)
  replacement_indices: Sequence[int] = attrs.field(init=True, kw_only=True)
  psr: PolynomialSparseRecovery = attrs.field(
      init=True, kw_only=True, factory=PolynomialSparseRecovery
  )

  def reset(self) -> None:
    raise NotImplementedError('Should not be used.')

  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    raise NotImplementedError('Should not be used.')

  def predict(self, x: np.ndarray) -> float:
    objectives = []
    for x_restrictor in self.X_restrictors:
      x_copy = np.copy(x)
      x_copy[self.replacement_indices] = x_restrictor[self.replacement_indices]
      objectives.append(self.psr.predict(x_copy))
    return np.mean(objectives)


def _binary_subset_enumeration(
    dim: int, indices: Sequence[int], default_value: float = 1.0
) -> np.ndarray:
  """Outputs all possible binary vectors from {-1, 1}^{dim} where only positions from `indices` are changed."""
  output = default_value * np.ones(
      shape=(2 ** len(indices), dim), dtype=np.float32
  )
  for i, binary in enumerate(
      itertools.product([-1.0, 1.0], repeat=len(indices))
  ):
    output[i, indices] = binary
  return output


@attrs.define
class HarmonicaQ(_Surrogate):
  """Q-stage Harmonica.

  At each stage:
  1. Invoke PSR on the previous data (X,Y).
  2. Obtain t maximizers of the surrogate of the PSR.
  3. Redefine a 'restricted surrogate' using the t maximizers.
  4. Produce a new (X', Y') dataset via random search on this 'restricted
  surrogate'.
  """

  # PolynomialSparseRecovery regressor.
  _psr: PolynomialSparseRecovery = attrs.field(
      init=True, kw_only=True, factory=PolynomialSparseRecovery
  )
  # Number of stages.
  _q: int = attrs.field(
      default=10, validator=attrs.validators.ge(0), kw_only=True
  )

  # Number of maximizers on the surrogate to use.
  _t: int = attrs.field(
      default=1, validator=attrs.validators.ge(1), kw_only=True
  )
  # Number of data samples to collect on the restricted surrogate.
  _T: int = attrs.field(
      default=300, validator=attrs.validators.ge(1), kw_only=True
  )

  _restricted_surrogate: Optional[RestrictedSurrogate] = None

  def __attrs_post_init__(self):
    self.reset()

  def reset(self) -> None:
    self._restricted_surrogate = None
    self._psr.reset()

  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    """Performs q-stage Harmonica."""
    num_vars = X.shape[-1]

    X_temp = X
    Y_temp = Y
    for _ in range(self._q):
      # Invoke PSR on data.
      self._psr.reset()
      self._psr.regress(X_temp, Y_temp)
      J = self._psr.index_set()

      # Perform brute force maximization to optain top t optimizers.
      all_X_in_J = _binary_subset_enumeration(num_vars, list(J))
      all_Y_in_J = np.array([self._psr.predict(x) for x in all_X_in_J])
      maximizer_idxs = np.argsort(all_Y_in_J)[-self._t :]
      X_maximizers = all_X_in_J[maximizer_idxs]

      # Define restricted surrogate and obtain data from it.
      self._restricted_surrogate = RestrictedSurrogate(
          X_restrictors=X_maximizers,
          replacement_indices=list(J),
          psr=self._psr,
      )
      X_temp = np.random.choice([-1.0, 1.0], size=(self._T, num_vars))
      Y_temp = np.array([self._restricted_surrogate.predict(x) for x in X_temp])

  def predict(self, x: np.ndarray) -> float:
    if self._restricted_surrogate is None:
      raise ValueError('You must call regress() first.')
    return self._restricted_surrogate.predict(x)


class HarmonicaDesigner(vza.Designer):
  """Harmonica Designer.

  The summary of the current implementation is as follows:

  1. Use previous trials for data collection.
  2. Perform Polynomial Lasso over the data and obtain a predictor function
  based on filtering only the highest coefficients.
  3. Obtain the "J"-set (i.e. set of input indices that only affect the
  predictor function).
  4. Find the global optimizer over the J-set domain on the predictor function,
  and return it as a suggestion.
  """

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      harmonica_q: Optional[HarmonicaQ] = None,
      acquisition_samples: int = 100,
      num_init_samples: int = 10,
  ):
    """Init.

    Args:
      problem_statement: Must use a boolean search space.
      harmonica_q: HarmonicaQ class. If None, will use default class with
        default kwargs.
      acquisition_samples: Number of trial samples to optimize final acquisition
        function.
      num_init_samples: Number of initial random suggestions for seeding the
        model.
    """

    if problem_statement.search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.'
      )
    for p_config in problem_statement.search_space.parameters:
      if p_config.external_type != vz.ExternalType.BOOLEAN:
        raise ValueError('Only boolean search spaces are supported.')

    self._problem_statement = problem_statement
    self._metric_name = self._problem_statement.metric_information.item().name
    self._search_space = problem_statement.search_space
    self._num_vars = len(self._search_space.parameters)

    self._harmonica_q = harmonica_q or HarmonicaQ()
    self._acquisition_samples = acquisition_samples
    self._num_init_samples = num_init_samples

    self._trials = []

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    self._trials += tuple(completed.trials)

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Performs entire q-stage Harmonica using previous trials for regression data.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1
    if count > 1:
      raise ValueError('This designer does not support batched suggestions.')

    if len(self._trials) < self._num_init_samples:
      random_designer = random.RandomDesigner(self._search_space)
      return random_designer.suggest(count)

    # Convert previous trial data into regression data.
    X = []
    Y = []
    for t in self._trials:
      single_x = [
          1.0 if t.parameters.get_value(p.name) == 'True' else -1.0
          for p in self._search_space.parameters
      ]
      single_y = t.final_measurement.metrics[self._metric_name].value
      X.append(single_x)
      Y.append(single_y)
    X = np.array(X)
    Y = np.array(Y)

    if (
        self._problem_statement.metric_information.item().goal
        == vz.ObjectiveMetricGoal.MINIMIZE
    ):
      Y = -Y

    # Perform q-stage Harmonica.
    self._harmonica_q.reset()
    self._harmonica_q.regress(X, Y)

    # Optimize final acquisition function.
    # TODO: Allow any designer instead of just random search.
    X_temp = np.random.choice(
        [-1.0, 1.0], size=(self._acquisition_samples, self._num_vars)
    )
    Y_temp = np.array([self._harmonica_q.predict(x) for x in X_temp])
    x_new = X_temp[np.argmax(Y_temp)]

    parameters = vz.ParameterDict()
    for i, p in enumerate(self._search_space.parameters):
      parameters[p.name] = 'True' if x_new[i] == 1.0 else 'False'
    return [vz.TrialSuggestion(parameters=parameters)]



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""High-level wrappers for stochastic process hyperparameter optimizers."""

import functools
from typing import Generic, Optional, Protocol, TypeVar

from absl import logging
import chex
import jax
from jax import numpy as jnp
from vizier._src.jax import stochastic_process_model as sp


Params = TypeVar('Params', bound=chex.ArrayTree)


class LossFunction(Protocol, Generic[Params]):
  """Evaluates model params and returns (loss, dict of auxiliary metrics)."""

  def __call__(self, params: Params) -> tuple[jax.Array, chex.ArrayTree]:
    """Evaluates model params and returns (loss, dict of auxiliary metrics)."""
    pass


class LossAndGradFunction(Protocol, Generic[Params]):
  """Computes loss and gradient."""

  def __call__(
      self, params: Params
  ) -> tuple[tuple[jax.Array, chex.ArrayTree], jax.Array]:
    """Returns (loss, tree of auxiliary metrics), grad."""


class Optimizer(Protocol[Params]):
  """Optimizes the LossFunction.

  Example:

  ```python
  setup: Setup = lambda rng: jax.random.uniform(rng, minval=-5, maxval=5)

  def loss_fn(xs):  # satisfies `LossFunction` Protocol
    xs = jax.nn.sigmoid(xs)
    return jnp.cos(xs * 5 * 2 * jnp.pi) * (2 - 5 * (xs - 0.5)**2), dict()

  optimize = optimizers.OptaxTrain(optax.adam(5e-3), epochs=500, verbose=True)
  rngs = jax.random.split(jax.random.PRNGKey(0), 51)
  optimal_params, metrics = optimize(
      init_params=jax.vmap(setup)(rngs[1:]),
      loss_fn=loss_fn,
      rng=rngs[0])
  ```
  """

  def __call__(
      self,
      init_params: Params,
      loss_fn: LossFunction[Params],
      rng: jax.Array,
      *,
      constraints: Optional[sp.Constraint] = None,
      best_n: Optional[int] = None,
  ) -> tuple[Params, chex.ArrayTree]:
    """Optimizes a LossFunction expecting Params as input.

    When constraint bijectors are applied, note that the returned parameters are
    in the constrained space (the parameter domain), not the unconstrained space
    over which the optimization takes place.

    If the Optimizer uses lower and upper bounds, then it is responsible for
    converting `None` bounds to `+inf` or `-inf` as necessary.

    Args:
      init_params: A set of initial points represented as a batched Pytree. All
        leaf nodes are expected to have the same leading dimension, which is the
        batch size.
      loss_fn: Evaluates a point.
      rng: JAX PRNGKey.
      constraints: Parameter constraints.
      best_n: If not None, returns a pytree with a batch dimension [best_n].

    Returns:
      Tuple containing optimal input in the constrained space and optimization
      metrics.
    """


def get_best_params(
    losses: jax.Array,
    all_params: chex.ArrayTree,
    *,
    best_n: Optional[int] = None,
) -> chex.ArrayTree:
  """Returns the top `best_n` parameters that minimize the losses.

  Args:
    losses: Shape (N,) array
    all_params: ArrayTree whose leaves have shape (N, ...)
    best_n: Integer greater than or equal to 1. If None, squeezes the leading
      dimension.

  Returns:
    Top `best_n` parameters.
  """
  argsorted = jnp.argsort(losses)
  if not best_n:
    best_idx = argsorted[:1]
  else:
    best_idx = argsorted[:best_n]

  logging.info('Best loss(es): %s at retry %s', losses[best_idx], best_idx)
  optimal_params = jax.tree_util.tree_map(lambda p: p[best_idx], all_params)
  if best_n is None:
    optimal_params = jax.tree.map(
        functools.partial(jnp.squeeze, axis=0), optimal_params
    )
  return optimal_params



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for random_sample."""

import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.random import random_sample

from absl.testing import absltest
from absl.testing import parameterized


class RandomSampleTest(parameterized.TestCase):

  def setUp(self):
    super(RandomSampleTest, self).setUp()
    self.rng = np.random.default_rng(0)

  @parameterized.named_parameters(
      dict(testcase_name='prob=0', value=0.2, target=0.0),
      dict(testcase_name='prob=1.0', value=-0.5, target=0.0),
      dict(testcase_name='prob=1.5', value=0.6, target=1.0),
      dict(testcase_name='prob=-0.1', value=3.0, target=2.0),
  )
  def test_get_closest_element(self, value, target):
    items = [0.0, 1.0, 2.0]
    self.assertEqual(random_sample.get_closest_element(items, value), target)

  def test_shuffle_list(self):
    items = ['a', 'b', 'c', 'd']
    shuffled_items = random_sample.shuffle_list(self.rng, items)
    # Check that all items appear once
    items_set = set(items)
    for item in shuffled_items:
      self.assertIn(item, items_set)
      items_set.remove(item)

  def test_sample_uniform(self):
    value1 = random_sample.sample_uniform(self.rng)
    self.assertTrue(0.0 <= value1 <= 1.0)
    value2 = random_sample.sample_uniform(self.rng)
    self.assertNotEqual(value1, value2)
    value3 = random_sample.sample_uniform(self.rng, 10.0, 20.0)
    self.assertTrue(10.0 <= value3 <= 20.0)

  def test_sample_bernoulli(self):
    value1 = random_sample.sample_bernoulli(self.rng, 0.5)
    self.assertIn(value1, [0, 1])
    value2 = random_sample.sample_bernoulli(self.rng, 0.0)
    self.assertEqual(value2, 1)
    value3 = random_sample.sample_bernoulli(self.rng, 1.0)
    self.assertEqual(value3, 0)
    value4 = random_sample.sample_bernoulli(self.rng, 0.5, 'A', 'B')
    self.assertIn(value4, ['A', 'B'])

  def test_sample_integer(self):
    value1 = random_sample.sample_integer(self.rng, 5, 5)
    self.assertEqual(value1, 5)
    value2 = random_sample.sample_integer(self.rng, 0, 10)
    self.assertIn(value2, list(range(0, 11)))

  def test_sample_categorical(self):
    value1 = random_sample.sample_categorical(self.rng, ['A'])
    self.assertEqual(value1, 'A')
    value2 = random_sample.sample_categorical(self.rng, ['A', 'B', 'C'])
    self.assertIn(value2, ['A', 'B', 'C'])

  def test_sample_discrete(self):
    value1 = random_sample.sample_discrete(self.rng, [2.5])
    self.assertEqual(value1, 2.5)
    value2 = random_sample.sample_discrete(self.rng, [0.0, 1.0, 10.0])
    self.assertIn(value2, [0.0, 1.0, 10.0])

  def test_sample_value_integer(self):
    int_param = vz.ParameterConfig.factory('int', bounds=(0, 10))
    value = random_sample._sample_value(self.rng, int_param)
    self.assertIn(value, list(range(0, 11)))

  def test_sample_value_discrete(self):
    discrete_param = vz.ParameterConfig.factory(
        'discrete', feasible_values=[1.0, 2.0, 10.0])
    value = random_sample._sample_value(self.rng, discrete_param)
    self.assertIn(value, [1.0, 2.0, 10.0])

  def test_sample_value_float(self):
    float_param = vz.ParameterConfig.factory('float', bounds=(0.0, 1.0))
    value = random_sample._sample_value(self.rng, float_param)
    self.assertTrue(0.0 <= value <= 1.0)

  def test_sample_value_categorical(self):
    categorical_param = vz.ParameterConfig.factory(
        'categorical', feasible_values=['a', 'b', 'c'])
    value = random_sample._sample_value(self.rng, categorical_param)
    self.assertIn(value, ['a', 'b', 'c'])

  def test_sample_input_parameters(self):
    space = vz.SearchSpace()
    root = space.root
    root.add_bool_param('b1')
    root.add_discrete_param('d1', [1.0, 2.0, 10.0])
    root.add_float_param('f1', 0.0, 15.0, scale_type=vz.ScaleType.LINEAR)
    root.add_float_param('f2', 100.0, 200.0, scale_type=vz.ScaleType.LINEAR)
    root.add_int_param('i1', 0, 10, scale_type=vz.ScaleType.LINEAR)
    root.add_categorical_param('c1', ['a', 'b', 'c'])
    parameter_dict = random_sample.sample_parameters(self.rng, space)
    self.assertIn(parameter_dict['b1'].value, ['True', 'False'])
    self.assertIn(parameter_dict['d1'].value, [1.0, 2.0, 10.0])
    self.assertTrue(0.0 <= parameter_dict['f1'].value <= 15.0)
    self.assertTrue(100.0 <= parameter_dict['f2'].value <= 200.0)
    self.assertIn(parameter_dict['i1'].value, list(range(0, 11)))
    self.assertIn(parameter_dict['c1'].value, ['a', 'b', 'c'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters.synthetic import branin
from vizier._src.benchmarks.testing import experimenter_testing

from absl.testing import absltest


class Branin2DExperimenterTest(absltest.TestCase):

  def test_branin_impl(self):
    np.testing.assert_allclose(
        branin._branin(
            np.array([[-np.pi, 12.275], [np.pi, 2.275], [9.42478, 2.475]])
        ),
        0.397887,
        atol=1e-5,
    )

  def test_experimenter_argmin(self):
    trial = vz.Trial(parameters={'x1': -np.pi, 'x2': 12.275})
    branin.Branin2DExperimenter().evaluate([trial])
    self.assertAlmostEqual(
        trial.final_measurement_or_die.metrics.get_value('value', np.nan),
        0.397887,
        places=5,
    )

  def test_experimenter(self):
    experimenter_testing.assert_evaluates_random_suggestions(
        self, branin.Branin2DExperimenter()
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for `transfer_learning.py`."""

import chex
import jax
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.designers.gp import transfer_learning as vtl

from absl.testing import absltest


tfd = tfp.distributions


class GoogleGpBanditTest(absltest.TestCase):

  def test_sequential_combine_predictions(self):
    prior_pred = vtl.TransferPredictionState(
        pred=tfd.Normal(loc=[0.1, 0.2], scale=[10.0, 5.0]),
        aux={},
        training_data_count=10,
        num_hyperparameters=5,
    )  # Try one case with `training_data_count` > `num_hyperparameters`.

    top_pred = vtl.TransferPredictionState(
        pred=tfd.Normal(loc=[0.1, 0.2], scale=[10.0, 5.0]),
        aux={},
        training_data_count=10,
        num_hyperparameters=15,
    )  # Try one case with `training_data_count` < `num_hyperparameters`.

    comb_pred, comb_aux = vtl.combine_predictions_with_aux(
        top_pred=top_pred, base_pred=prior_pred
    )

    # The sum of means should be precisely `0.1 + 0.1` and `0.2 + 0.2`.
    self.assertEqual(list(comb_pred.mean()), [0.2, 0.4])

    # The combination of stddevs should be approximately the same as each
    # individual standard deviation, because we are combining the same
    # standard deviations together.
    self.assertSequenceAlmostEqual(comb_pred.stddev(), [10.0, 5.0], places=4)

    # Batch shapes in aux must be the same as the batch shape of predictions
    chex.assert_tree_shape_prefix(comb_aux, [2])


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  jax.config.update('jax_enable_x64', True)
  jax.config.update('jax_log_compiles', True)
  absltest.main()
